%% smooth_dem_robust_bilinear
%
% reconstruct surface with 
%   energy = sum_m rho(l_m-a_m)/sigma_n) + 
%             1/sigma_d sum_ij d_ii^2+2d_ij^2+djj^2
% l_m = observed heights, interpolated bilinearly
% a_m unknown grid heights, d_ij second derivatives
% 
% function [dem_smoothed,S,Sigma,Np,Nr,Mc,v,A,weights,W] = ...
%    smooth_dem_robust_bilinear...
%    (points,BB,delta_x,sigma_k,out_C,out_print,type_robust,type_data)
%
% points  = Nx4-matrix of (x,y,h,sigma) values, 
% BB      = [xmin,ymin,xmax,ymax] boundin,g box of square grid
% delta_x = grid size, squared 
%           xmax, ymax will be adapted, not decreased
%           points not incide BB will be discarded
% sigms_k = std. dev. of curvature of surface, referring to delta_x
% out_C   = 0 no covariance matrix as output, (default if U>1600)
%           1 covariance matrix as output
% out_print = 0 no output
%             1 little output
%             2 much output
% type_robust = 0,1,2,3 (00,01,10,11) for points and dem
%
% dem_smoothe = Nr x Mc matrix of smoothed heigths
% S           = corresponding standard deviations, if required
% Sigma       = full covariance matrix. if required
% Np          = number of points
% Nr, Mc      = number of grid points in x- and y-direction
% v           = residuals (Np + (Nr-2)*Mc + Nr*(Mc-2) + (Nr-1)*(Mc-1)) for
%               points
%               second derivatives cc
%               second derivatives rr
%               torsion
% A           = design matrix (sparse) 
% weights     = weight-factors after robust estimation (0 or 1)
% W           = weight of observations including both, uncertainty and robust factor 
%
% grid is matrix Nr x Mc and stored column wise
%
% wf 7/2014
%
 
function [dem_smoothed,S,Sigma,Np,Nr,Mc,v,A,weights,weights_f,W] = ...
    smooth_dem_robust_bilinear...
    (points,BB,delta_x,sigma_k,out_C,type_robust,type_data,...
    out_in,print_type,plot_type)
    

%% number of points, preliminary --> final (which are in BB)
[Npp,tmp] = size(points);

xmin = BB(1);
ymin = BB(2);
xmax = BB(3);
ymax = BB(4);
Nr =ceil((xmax-xmin)/delta_x)+1;
Mc =ceil((ymax-ymin)/delta_x)+1;


Np = 0;
for p=1:Npp
    if points(p,1) >= xmin && ...
            points(p,1) < xmax && ...
            points(p,2) >= ymin && ...
            points(p,2) < ymax
        Np=Np+1;
        poi(Np,:) = [(points(p,1)-xmin)/delta_x,(points(p,2)-ymin)/delta_x,points(p,3:4)];
        
    end
end
            
%% number of unknowns and observations incl. prior
U = Nr*Mc;
N = Np + (Nr-2)*Mc + Nr*(Mc-2) + (Nr-1)*(Mc-1);
            if print_type > 0
                number_of_unknowns = U %#ok<NOPRT,NASGU>
            end

%% do not give covariance matrix for U>10000.
            if U > 40401
                out_C = 0;
            end
            if print_type > 0 
                tic
            end
%% estimate iteratively

iter_switch=3;
if type_robust == 0
    N_iter = 1;
else
    N_iter=6;
end

xa = zeros(U,1);               % initial unknown parameters
weights   = ones(N,1);         % weight-factors for robust estimation
weights_f = weights;
sigma_k0  = sigma_k;
gain=1.0;

for iter = 1:N_iter
    L1 = 0;
    if type_robust > 0 && iter <= iter_switch
        L1=1;
    else
        L1=0;
    end
    sigma_k=sigma_k0*gain^(N_iter-iter);     % adapt prior sigma
    
    [A,weights,dx,v,W,Nmatrix]=estimate_dem_robust...
        (N,U,Np,Nr,Mc,poi,xa,weights,sigma_k,iter,type_robust,L1,...
        out_in,print_type,plot_type);
    
    xa = xa + dx;
     
    if print_type > 0
        iter_dx = [iter,norm(dx)] %#ok<NOPRT,NASGU>
    end
    %%
                if plot_type > 0
                    if iter == 1
                        if type_robust > 0
                        figure
                        hold on
                        title('surf')
                        X=([4*Nr:4*Nr+Nr-1]'*ones(1,Mc))*delta_x+BB(1);
                        Y=(ones(Nr,1)*[0:Mc-1])*delta_x+BB(2);

                            if type_data ~=5 
                                colormap(gray)
                                surf(X,Y,reshape(xa,Nr,Mc));
                                alpha(0.3);
                                %mesh(ds);
                                view([-29,65])
                                plot3(points(:,1)+4*Nr*delta_x,points(:,2),points(:,3),'.r','MarkerSize',3);

                            else
                                subplot(1,3,2)
                                %imagesc(reshape(xa,Nr,Mc));
                                imshow(reshape(xa,Nr,Mc));
                                colormap(gray)
                                view(0,90)
                                title('1.iteration')
                            end
                            title(strcat('iteration=',num2str(iter),' fitted dem'))

                            axis equal
                        end
                    else
                        figure
                        subplot(2,2,1)
                        hold on
                        X=([0:Nr-1]'*ones(1,Mc))*delta_x+BB(1);
                        Y=(ones(Nr,1)*[0:Mc-1])*delta_x+BB(2);
                        colormap(gray)
                        mesh(X,Y,reshape(xa,Nr,Mc),'EdgeColor',[0,0,0])%  %mesh(ds,'EdgeColor',[0,0,0])
                        %axis equal
                        view([-29,65])
                        zlim([0,60])
                        grid off
                        %alpha(0.3);
                        %mesh(ds);
                        %plot3(points(:,1),points(:,2),points(:,3),'.r','MarkerSize',5);

                        title(strcat('iteration=',num2str(iter),' fitted dem'))

                        if type_data ~= 5
                            axis equal
                        end
                        Icc = Np;                    % first index for column curvatures
                        Irr = Np + (Nr-2)*Mc;        % first index for row curvature
                        Irc = Np + (Nr-2)*Mc + Nr*(Mc-2);    % first index for torsion
                        subplot(2,2,2)
                        imagesc(reshape(weights(Icc+1:Irr),Nr-2,Mc));
                        subplot(2,2,3)
                        if type_robust ~= 2
                            imagesc(reshape(weights(Irr+1:Irc),Nr,Mc-2));
                        else
                            rr=floor(sqrt(Np));
                            imagesc(reshape(1-weights(1:rr^2),rr,rr))
                        end
                        subplot(2,2,4)
                        if type_robust ~= 2
                            imagesc(reshape(weights(Irc+1:end),Nr-1,Mc-1));
                        else
                            rr=floor(sqrt(Np));
                            imagesc(reshape(out_in(1:rr^2),rr,rr))
                        end
                    end
                end


end
if type_robust > 0
    %final estimate
    weights_f = 0.9999*ceil(weights-0.5)+0.0001;
    if print_type > 1
        weights_final = weights_f(1:20)' %#ok<NOPRT,NASGU>
    end
    xa=zeros(length(xa),1);
    
    [A,weights_f,dx,v,W,Nmatrix]=estimate_dem_robust...
        (N,U,Np,Nr,Mc,poi,xa,weights_f,sigma_k,1,0,0,...
        out_in,print_type,plot_type);
    
    xa = xa + dx;
                if print_type > 0
                    iter_dx = [iter+1,norm(dx)] %#ok<NOPRT,NASGU>
                end

                if plot_type > 0
                figure
                    subplot(2,2,1)
                    hold on
                    X=([0:Nr-1]'*ones(1,Mc))*delta_x+BB(1);
                    Y=(ones(Nr,1)*[0:Mc-1])*delta_x+BB(2);
                    colormap(gray)
                    mesh(X,Y,reshape(xa,Nr,Mc),'EdgeColor',[0,0,0])%  %mesh(ds,'EdgeColor',[0,0,0])
                    %axis equal
                    view([-29,65])
                    zlim([0,60])
                    grid off
                    %alpha(0.3);
                    %mesh(ds);
                    %plot3(points(:,1),points(:,2),points(:,3),'.r','MarkerSize',5);

                    title(strcat('last iteration, fitted dem'))

                    if type_data ~= 5
                        axis equal
                    end
                    Icc = Np;                    % first index for column curvatures
                    Irr = Np + (Nr-2)*Mc;        % first index for row curvature
                    Irc = Np + (Nr-2)*Mc + Nr*(Mc-2);    % first index for torsion
                    subplot(2,2,2)
                    imagesc(reshape(weights_f(Icc+1:Irr),Nr-2,Mc));
                    subplot(2,2,3)
                    if type_robust ~= 2
                        imagesc(reshape(weights(Irr+1:Irc),Nr,Mc-2));
                    else
                       rr=floor(sqrt(Np));
                       imagesc(reshape(1-weights(1:rr^2),rr,rr))
                    end
                    subplot(2,2,4)       
                    if type_robust ~= 2
                       imagesc(reshape(weights(Irc+1:end),Nr-1,Mc-1));
                    else
                       rr=floor(sqrt(Np));
                       imagesc(reshape(out_in(1:rr^2),rr,rr))
                    end
                end
end

dem_smoothed = reshape(xa,Nr,Mc);

%% determine covariance matrix for quality analysis
Sigma=0;
S = 0;

if out_C > 0    
                  
    Sigma = sparseinv(Nmatrix);       % use sparse inverse    
                
    S     = full(reshape(diag(Sigma),Nr,Mc));
end



return